import 'dart:async';

import 'package:logging/logging.dart';
import 'package:shared_preferences/shared_preferences.dart';

import 'environment_config.dart';
import 'environment_type.dart';

final log = Logger('EnvironmentManager');

class EnvironmentManager {
  /// Return singleton object
  factory EnvironmentManager() {
    return _instance;
  }

  EnvironmentManager._internal();

  static final EnvironmentManager _instance = EnvironmentManager._internal();

  Map<String, EnvironmentConfig> _configMap = {};
  Map<String, EnvironmentType> _environmentTypeMap = {};
  bool isNeedRestart = true;
  EnvironmentType defaultEnvironmentType = EnvironmentType.product;

  StreamController<Map<String, EnvironmentType>> _environmentChangedStreamController = StreamController<Map<String, EnvironmentType>>.broadcast();

  Stream<Map<String, EnvironmentType>> get environmentChangedListener {
    return _environmentChangedStreamController.stream;
  }

  void dispose() {
    _environmentChangedStreamController.close();
  }

  void registerEnvironmentConfig(EnvironmentConfig config) {
    if (_configMap[config.name] == null) {
      log.info('Register environment: ${config.name} ${config.displayName}');
      _configMap[config.name] = config;
      _environmentTypeMap[config.name] = EnvironmentType.product;
    } else {
      log.warning('Register environment duplicate name');
    }
  }

  Future<void> loadEnvironmentConfig() async {
    SharedPreferences preferences = await SharedPreferences.getInstance();
    await preferences.setBool('app_environment_restart', true);
    for (final name in _configMap.keys) {
      _environmentTypeMap[name] = EnvironmentType.values[preferences.getInt('app_environment_$name') ?? defaultEnvironmentType.index];
    }
    isNeedRestart = true;
  }

  Future<void> saveEnvironmentConfig() async {
    SharedPreferences preferences = await SharedPreferences.getInstance();
    for (final name in _environmentTypeMap.keys) {
      await preferences.setInt('app_environment_$name', _environmentTypeMap[name].index);
    }
    await preferences.setBool('app_environment_restart', false);
    isNeedRestart = false;
  }

  void setEnvironment(EnvironmentType type, [String name]) {
    if (name != null && name.isNotEmpty) {
      if (_environmentTypeMap[name] != null) {
        _environmentTypeMap[name] = type;
      }
    } else {
      for (final name in _configMap.keys) {
        _environmentTypeMap[name] = type;
      }
    }
    _environmentChangedStreamController.add(_environmentTypeMap);
    saveEnvironmentConfig();
  }

  Map<String, EnvironmentType> get currentEnvironment {
    bool isAllSame = true;
    EnvironmentType environmentType = _environmentTypeMap[_environmentTypeMap.keys.first];
    for (final name in _environmentTypeMap.keys) {
      if (environmentType != _environmentTypeMap[name]) {
        isAllSame = false;
        break;
      }
    }

    if (isAllSame) {
      return {'all': environmentType};
    }

    return _environmentTypeMap;
  }

  String _getConfig(EnvironmentType environmentType, String name, String key) {
    EnvironmentConfig config = _configMap[name];
    if (config != null) {
      return config.getConfig(environmentType, key);
    }
    return null;
  }

  String getConfig(String name, String key) {
    List<EnvironmentType> environmentOrder = [
      EnvironmentType.product,
      EnvironmentType.gray,
      EnvironmentType.test,
      EnvironmentType.develop,
      EnvironmentType.benchmark,
    ];

    EnvironmentType environmentType = _environmentTypeMap[name];
    if (environmentType != null) {
      int environmentIndex = environmentOrder.indexOf(environmentType);

      String result;
      while (result == null && environmentIndex >= 0) {
        result = _getConfig(environmentOrder[environmentIndex], name, key);
        environmentIndex -= 1;
      }

      return result;
    }

    return null;
  }

  void reset() {
    _configMap = {};
    _environmentTypeMap = {};
  }
}
